module("L_DLNAUPnPUtility", package.seeall)

local url = require("socket.url")
local socket = require("socket")
local http = require("socket.http")
local ltn12 = require("ltn12")

-- 5 Second timeout
http.TIMEOUT = 5

local UPNP_DISCOVERY = "M-SEARCH * HTTP/1.1\r\nHOST: 239.255.255.250:1900\r\nMAN: \"ssdp:discover\"\r\nMX: 5\r\nST: %s\r\n\r\n"

local UPNP_REQUEST = [[<?xml version="1.0" encoding="utf-8"?>
<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
<s:Body>
<u:%s xmlns:u="%s">%s</u:%s>
</s:Body>
</s:Envelope>]]

local DIDL_FORMAT=[[<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">
<item id="%s" parentID="%s" restricted="true">
<res protocolInfo="x-file-cifs:*:audio/mpeg:*">%s</res>
<r:streamContent>%s</r:streamContent>
<upnp:albumArtURI>%s</upnp:albumArtURI>
<dc:title>%s</dc:title>
<upnp:class>%s</upnp:class>
<dc:creator>%s</dc:creator>
<upnp:album>%s</upnp:album>
<upnp:originalTrackNumber>%s</upnp:originalTrackNumber>
<r:albumArtist>%s</r:albumArtist>
</item>
</DIDL-Lite>
]]

local VeraIP = ""
local log = print
local error = print
local contentType
local Services = {}

function initialize(veraIP, logger, errorLogger, ct)
  VeraIP = veraIP
  log = logger
  error = errorLogger
  contentType = ct or 'text/xml; charset="utf-8"'
end

function decode(val)
      return val:gsub("&amp;", "&")
                :gsub("&lt;", "<")
                :gsub("&gt;", ">")
                :gsub("&quot;", '"')
                :gsub("&#60;", '<')
                :gsub("&#62;", '>')
                :gsub("&#34;", '"')
end

local function encode(val)
      return val:gsub("&", "&amp;")
                :gsub("<", "&lt;")
                :gsub(">", "&gt;")
                :gsub('"', "&quot;")
end

function unformatXML(val)
      return val:gsub("^%s+<", "<")
                :gsub(">%s+$", ">")
                :gsub(">%s+<", "><")
end

function createDIDL(resURI, creator, album, title, artist, streamContent, albumArtURI, trackNumber, class, id, parentID)
  return DIDL_FORMAT:format(encode(id or 1),
                            encode(parentID or -1),
                            encode(resURI or ""),
                            encode(streamContent or ""),
                            encode(albumArtURI or ""),
                            encode(title or "No title"),
                            encode(class or "object.item.audioItem.musicTrack"),
                            encode(creator or ""),
                            encode(album or ""),
                            encode(trackNumber or 1),
                            encode(artist or "No artist"))
end

-- Send a UPnP SUBSCRIBE to the UPnP device,
-- Return value on success:
--   SID (subscription ID) provided by the device.
--   Duration (in seconds) before the subscription must be renewed
-- Return value on failure:
--   nil
--   Reason for failure (string)
-- Code coming from Wemo plugin (function subscribeToDevice) and adjusted
function UPnP_subscribe(ipAddress, ipPort, serverURL, callbackURL, renewalSID)
	local eventSubURL = serverURL:format(ipAddress, ipPort)

	-- Create a socket with a timeout.
	local sock = function()
		local s = socket.tcp()
		s:settimeout(http.TIMEOUT)
		return s
	end

	local headers = {
		["TIMEOUT"] = "Second-3600",
	}

	if (renewalSID) then
		-- Renewing, include SID header.
		headers["SID"] = renewalSID
	else
		-- New subscription, include CALLBACK and NT headers
		headers["CALLBACK"] = string.format("<%s>", callbackURL)
		headers["NT"] = "upnp:event"
	end

	-- Ask the device to inform the proxy about status changes.
	local request, code, headers = http.request({
		url = eventSubURL,
		method = "SUBSCRIBE",
		headers = headers,
		create = sock,
	})

	if (request == nil and code ~= "closed") then
		log("Failed to subscribe to " .. eventSubURL .. ": " .. code)
		return nil, code
	elseif (code ~= 200) then
		log("Failed to subscribe to " .. eventSubURL .. ": " .. code)
		return nil, code
	else
		local duration = headers["timeout"]:match("Second%-(%d+)")
		log("Subscription confirmed, SID = " .. headers["sid"] .. " with timeout " .. duration)
		return headers["sid"], tonumber(duration)
	end
end
  
  
function UPnP_request(ipAddress, ipPort, serverURL, action, servicetype, args)
  local function table2XML(value)
    local result = ""

    --
    -- Convert all the Number, Boolean and Table objects, and escape all the string
    -- values in the XML output stream
    --
    -- If value table has an entry OrderedArgs, we consider that all the values
    -- are set in this special entry and we bypass all the other values of the value table.
    -- In this particular case, this entry itself is a table of strings, each element of
    -- the table following the format "parameter=value"
    --
    if (value.OrderedArgs ~= nil and type(value.OrderedArgs) == "table") then
        for i, val in ipairs(value.OrderedArgs) do
            local e, v = val:match("([^=]+)=(.*)")
            if (e ~= nil) then
                if (v == nil) then
                    result = result .. string.format("<%s />", e)
                elseif (type(v) == "table") then
                    result = result .. table2XML(v)
                elseif (type(v) == "number") then
                    result = result .. string.format("<%s>%.0f</%s>", e, v, e)
                elseif (type(v) == "boolean") then
                    result = result .. string.format("<%s>%s</%s>", e, (v and "1" or "0"), e)
                else
                    result = result .. string.format("<%s>%s</%s>", e, encode(v), e)
                end
            end
        end
    else
        for e, v in pairs(value) do
            if (v == nil) then
                result = result .. string.format("<%s />", e)
            elseif (type(v) == "table") then
                result = result .. table2XML(v)
            elseif (type(v) == "number") then
                result = result .. string.format("<%s>%.0f</%s>", e, v, e)
            elseif (type(v) == "boolean") then
                result = result .. string.format("<%s>%s</%s>", e, (v and "1" or "0"), e)
            else
                result = result .. string.format("<%s>%s</%s>", e, encode(v), e)
            end
        end
    end

    return result
  end

  local postBody = string.format(UPNP_REQUEST, action, servicetype, table2XML(args), action)
  serverURL = serverURL:format(ipAddress, ipPort)
  log(string.format("UPnP_request: url=[%s], body=[%s]", serverURL, postBody))

  --
  -- Execute the resulting URL, and collect the results as a Table
  --
  local resultTable = {}
  local status, statusMsg = http.request{
    url = serverURL,
    sink = ltn12.sink.table(resultTable),
    method = "POST",
    headers = {["Accept"] = "*/*",
               ["SOAPAction"] = '"' .. servicetype .. "#" .. action .. '"',
               ["Connection"] = "close",
               ["Content-Length"] = postBody:len(),
               ["Content-Type"] = contentType},
    source = ltn12.source.string(postBody),
  }

  --
  -- Flatten the resultTable into a regular string
  --
  local data = ""
  for i, v in ipairs(resultTable) do
    data = data .. v
  end
  
  if (status == nil) then
    --
    -- Handle TIMEOUT
    --
    return false, statusMsg or "An error occurred during the UPnP call"
  end
  
  if (tostring(statusMsg) == "200") then
    --
    -- Handle SUCCESS
    --
    log(string.format("UPnP_request: status=%s statusMsg=%s result=[%s]", status or "no status",
                      statusMsg or "no message",
                      data or "no result"))

    if (data == nil or data == "") then
      return true, ""
    else
      local pattern = string.format('.*<.-:Body><.-:%sResponse xmlns:.-="urn:.-">(.*)</.-:%sResponse></.-:Body></.-:Envelope>.*',
                                    action, action)
      local value = unformatXML(data):match(pattern) or ""

      -- TODO: Handle UPnP Error responses
      -- TODO: Handle Non-Scalar results

      -- commented out and returns a simple string instead. Due to multiple tags (not nested though)
      --      [[local dataTable = {}
      --      for tagBegin, value, tagEnd in value:gmatch("<(.*)>(.*)</(.*)>") do
      --        print(tagBegin .. " -> " .. value .. " <- " .. tagEnd)
      --        dataTable[tagBegin] = value
      --      end

      --     return dataTable]]
      return true, decode(value)
    end
  else
    --
    -- UNKNOWN ERROR
    --
    error(string.format("UPnP_request (%s, %s): status=%s statusMsg=%s result=[%s]",
                      action or "no action",
                      servicetype or "no servicetype",
                      status or "no status",
                      statusMsg or "no message",
                      data or "no result"))

    assert("Unhandled Response statusMsg=" .. statusMsg)
  end
end -- function UPnP_request

--
-- Return a handle to a UPnP Service object.
-- This can be used to call arbitrary Service Methods on that service using dynamic naming foo.
--
function service(ip, port, controlURL, servicetype, actions)
    local self = {}
    local mt = {}

    mt.__index = function(table, key)
        log("service.__index: Accessing non-existing function " .. key)

        local fn = function(...)
            if (actions[key]) then
                log(string.format("%s('%s', %d, '%s', '%s') Called with parameter count=%d", key, ip, port, controlURL, servicetype, select("#", ...)))
                return UPnP_request(ip, port, controlURL, key, servicetype, ...)
            else
                return false, "action not available"
            end
        end

        table[key] = fn
        return fn
    end

    setmetatable(self, mt)

    return self
end


--------------------------------------------------------------------------------
-- UPnP discovery
--------------------------------------------------------------------------------


function UPnP_discover(target)
    local devices = {}
    local udp = socket.udp()
    if (udp ~= nil) then
        local result, message = udp:sendto(UPNP_DISCOVERY:format(target), "239.255.255.250", 1900)
        if (result ~= nil) then
            udp:settimeout(5)
            for i=1,500 do
                result, message = udp:receivefrom()
                if (result == nil) then
                    break
                else
                    local location, ip, port = result:match("[Ll][Oo][Cc][Aa][Tt][Ii][Oo][Nn]:%s?(http://([%d%.]-):(%d+)/.-)\r\n")
                    local st = result:match("[Ss][Tt]:%s?(.-)\r\n")
                    if (location ~= nil and ip ~= nil and port ~= nil and st ~= nil) then
                        local new = true
                        for i,device in ipairs(devices) do
                            if (device.descriptionURL == location and device.st == st) then
                                new = false
                                break
                            end
                        end
                        if (new) then
                            table.insert(devices, { descriptionURL=location, ip=ip, port=port, st=st })
                        end
                    end
                end
            end
        end
        udp:close()
    end
    return devices
end


function scanUPnPDevices(deviceType, infos)
    local xml = "<devices>"
    local devices = UPnP_discover(deviceType)
    for i,dev in ipairs(devices) do
        local descrXML = UPnP_getDeviceDescription(dev.descriptionURL)
        if (descrXML ~= nil) then
            xml = xml .. "<device>"
            xml = xml .. "<ip>" .. (dev.ip or "") .. "</ip>"
            xml = xml .. "<port>" .. (dev.port or "") .. "</port>"
            xml = xml .. "<descriptionURL>" .. (dev.descriptionURL or "") .. "</descriptionURL>"
            local values = getInfosFromDescription(descrXML, deviceType, infos)
            if (infos ~= nil) then
                for i,tag in ipairs(infos) do
                    xml = xml .. "<" .. tag .. ">" .. (values[tag] or "") .. "</" .. tag .. ">"
                end
            end
            xml = xml .. "</device>"
        end
    end
    xml = xml .. "</devices>"
    return xml
end


function searchUPnPDevices(deviceType, name, ip)
    local devices = UPnP_discover(deviceType)
    for i,dev in ipairs(devices) do
        if (ip == nil or ip == "" or ip == dev.ip) then
            local descrXML = UPnP_getDeviceDescription(dev.descriptionURL)
            if (descrXML ~= nil) then
                local values = getInfosFromDescription(descrXML, deviceType, { "modelName" })
                if (values.modelName == name) then
                    return dev.descriptionURL
                end
            end
        end
    end
    return nil
end


--------------------------------------------------------------------------------
-- UPnP device and services description
--------------------------------------------------------------------------------


function UPnP_getDeviceDescription(descriptionURL)
    local status, xml = luup.inet.wget(descriptionURL, 2)
    if (status ~= 0) then
        xml = nil
    end
    return xml
end


function getInfosFromDescription(descriptionXML, deviceType, infos)
    local data = {}
    local tag, value, devices, device
    if (infos ~= nil) then
        local rootDevPart1, embeddedDevices, rootDevPart2 = descriptionXML:match(".-(<device>.-)<deviceList>(.-)</deviceList>(.-</device>).-")
        if (rootDevPart1 == nil or rootDevPart2 == nil or embeddedDevices == nil) then
            devices = descriptionXML:match(".-(<device>.*</device>).-") or ""
        elseif (deviceType ~= nil) then
            devices = rootDevPart1 .. rootDevPart2 .. embeddedDevices
        else
            devices = rootDevPart1 .. rootDevPart2
        end
        for device in devices:gmatch("<device>(.-)</device>") do
            value = device:match("<deviceType>(.+)</deviceType>")
            if (deviceType == nil or value == deviceType) then
                for i,tag in ipairs(infos) do
                    value = device:match("<"..tag..">(.+)</"..tag..">")
                    if (value ~= nil) then
                        data[tag] = value
                    end
                end
            end
        end
    end
    return data
end


function getIconFromDescription(descriptionXML)
    local resultURL = nil
    local rootDevice, icon, size, height, width, icnoURL
    local rootDevPart1, embeddedDevices, rootDevPart2 = descriptionXML:match(".-(<device>.-)<deviceList>(.-)</deviceList>(.-</device>).-")
    if (rootDevPart1 == nil or rootDevPart2 == nil or embeddedDevices == nil) then
        rootDevice = descriptionXML:match(".-(<device>.*</device>).-") or ""
    else
        rootDevice = rootDevPart1 .. rootDevPart2
    end
    size = 0
    for icon in rootDevice:gmatch("<icon>(.-)</icon>") do
        height = icon:match("<height>(.+)</height>")
        width = icon:match("<width>(.+)</width>")
        icnoURL = icon:match("<url>(.+)</url>")
        if (icnoURL ~= nil and height ~= nil and width ~= nil) then
            height = tonumber(height)
            width = tonumber(width)
            if (height >= width and height > size) then
                size = tonumber(height)
                resultURL = icnoURL
            elseif (height < width and width > size) then
                size = tonumber(width)
                resultURL = icnoURL
            end
        end
    end
    return resultURL
end


function getActionsFromSCPD(scpdURL)
    local actions = {}
    local action, name
    local status, xml = luup.inet.wget(scpdURL, 2)
    if (status == 0) then
        for action in xml:gmatch("<action>(.-)</action>") do
            name = action:match("<name>(.-)</name>")
            if (name ~= nil) then
                actions[name] = true
            end
        end
    end
    return actions
end


function getServicesFromDescription(descriptionXML, deviceType, baseURL, subsetServices)
    local services = {}
    local value, devices, device, service, service2, serviceType, serviceId, controlURL, eventSubURL
    if (subsetServices ~= nil) then
        local rootDevPart1, embeddedDevices, rootDevPart2 = descriptionXML:match(".-(<device>.-)<deviceList>(.-)</deviceList>(.-</device>).-")
        if (rootDevPart1 == nil or rootDevPart2 == nil or embeddedDevices == nil) then
            devices = descriptionXML:match(".-(<device>.*</device>).-") or ""
        elseif (deviceType ~= nil) then
            devices = rootDevPart1 .. rootDevPart2 .. embeddedDevices
        else
            devices = rootDevPart1 .. rootDevPart2
        end
        for device in devices:gmatch("<device>(.-)</device>") do
            value = device:match("<deviceType>(.+)</deviceType>")
            if (deviceType == nil or value == deviceType) then
                for service in device:gmatch("<service>(.-)</service>") do
                    serviceType = service:match("<serviceType>(.*)</serviceType>")
                    serviceId = service:match("<serviceId>(.*)</serviceId>")
                    controlURL = service:match("<controlURL>(.*)</controlURL>")
                    eventSubURL = service:match("<eventSubURL>(.*)</eventSubURL>")
                    scpdURL = service:match("<SCPDURL>(.*)</SCPDURL>")
                    if (serviceType ~= nil and serviceId ~= nil
                           and controlURL ~= nil and eventSubURL ~= nil
                           and scpdURL ~= nil) then
                        for i,service2 in ipairs(subsetServices) do
                            if (service2 == serviceType) then
                               services[serviceType] = {
                                       controlURL = url.absolute(baseURL, controlURL),
                                       eventSubURL = url.absolute(baseURL, eventSubURL),
                                       scpdURL = url.absolute(baseURL, scpdURL),
                                       serviceId = serviceId,
                                       actions = getActionsFromSCPD(url.absolute(baseURL, scpdURL)),
                                       object = nil }
                                break
                            end
                        end
                    end
                end
            end
        end
    end
    return services
end


--------------------------------------------------------------------------------
-- Services management
--------------------------------------------------------------------------------


  function aresServicesLoaded(uuid)
    if (uuid ~= nil and uuid ~= "" and Services[uuid] ~= nil) then
        return true
    else
        return false
    end
  end

  function getService(uuid, serviceType)
    local s = nil
    if (aresServicesLoaded(uuid) and Services[uuid][serviceType] ~= nil) then
        s = Services[uuid][serviceType].object
    end
    return s
  end

  function resetServices(uuid)
    if (aresServicesLoaded(uuid)) then
        Services[uuid] = nil
    end
  end

  function addServices(uuid, ip, port, descriptionXML, deviceType, subsetServices)
    if (uuid == nil or uuid == "") then
        return
    end

    if (Services[uuid] == nil) then
        Services[uuid] = {}
    end

    local services = getServicesFromDescription(
                              descriptionXML, deviceType,
                              string.format("http://%s:%s/", ip, port),
                              subsetServices)
    for k,v in pairs(services) do
        Services[uuid][k] = v
    end

    for k,v in pairs(Services[uuid]) do
        log(k .. " " .. v.serviceId .. " " .. v.controlURL .. " " .. v.eventSubURL .. " " .. v.scpdURL)
        if (v.object == nil) then
            v.object = service(ip, port, v.controlURL, k, v.actions)
        end
    end
  end



  function testURL(URLtoBeChecked)
    local sock = function()
        local s = socket.tcp()
        s:settimeout(2)
        return s
    end

    local t = {}
    local request, code = http.request({
        url = URLtoBeChecked,
        method = "HEAD",
        create = sock,
        sink = ltn12.sink.table(t)
    })

    if (request ~= nil and code == 200) then
        return true
    else
        return false
    end
  end


  --------------------------------------------------------------------------------
  -- UPnP AV metadata (DIDL-Lite parsing)
  --------------------------------------------------------------------------------


  -- parseElt(value, tag, subTag)
  -- Parse XML data at the second or third depth level
  -- If there are several elements, only the first is taken in consideration
  -- Parameters:
  --   value is the XML data
  --   tag is the first depth level XML tag
  --   subTag is the second depth level XML tag or nil if data have to be retrieved from the second depth level
  -- Return values:
  --   multilines string containing all parsed values; each line is formatted like that: tag="value"
  --   table containing parsed data (attributes + values) indexed either by the tag or tag@attribute
  function parseFirstElt(value, tag, subTag)
      local pattern, tmp, attributes, token, value1, attr, value2
      pattern = string.format("<%s(%%s?.-)>(.*)</%s>", tag, tag)
      attributes, tmp = value:match(pattern)
      if (tmp ~= nil and subTag ~= nil) then
          -- Get the content of the first sub tag
          pattern = string.format("<%s(%%s?.-)>(.-)</%s>.*", subTag, subTag)
          attributes, tmp = tmp:match(pattern)
      end

      if (tmp == nil) then
          return nil, {}
      end

      local elts, eltsTable = "", {}

      for attr, value1 in attributes:gmatch('%s(.-)="(.-)"') do
          eltsTable["@" .. attr] = value1
      end
	
      for token, attributes, value1 in tmp:gmatch("<([a-zA-Z0-9:]+)(%s?.-)>(.-)</.->") do
          elts = elts .. string.format('%s="%s"\n', token, value1)
          eltsTable[token] = value1
          for attr, value2 in attributes:gmatch('%s(.-)="(.-)"') do
              eltsTable[token .. "@" .. attr] = value2
          end
      end

      return elts, eltsTable
  end

  -- parseDIDLItem(value)
  -- Parse XML data for the item of a DIDL UPnP AV defining metadata XML meta data
  -- Parameters:
  --   value is the DIDL XML data
  -- Return values:
  --   multilines string containing all parsed values; each line is formatted like that: tag="value"
  --   table containing parsed data (attributes + values) indexed either by the tag or tag@attribute
  function parseDIDLItem(value)
      return parseFirstElt(value, "DIDL%-Lite", "item")
  end


  function getRes(value)
      local resTable = {}

      local tmp, protocol, res
      tmp = value:match("<DIDL%-Lite%s?.-><item%s?.->(.-)</item>.*</DIDL%-Lite>")
      if (tmp ~= nil) then
          for protocol, res in tmp:gmatch('<res%s?.-%s?protocolInfo="(.-)"%s?.->(.-)</res>') do
              resTable[protocol] = res
          end
      end

      return resTable
  end


  local function extractElement(tag, xml, default)
    if (xml == nil) then
      return(default)
    else
      local pattern = string.format("<%s%%s?.->(.*)</%s>", tag, tag)
      local result = (xml:match(pattern) or default)
      return result
    end
  end


  local function mergeXml(xml1, xml2)
    local result = ""
    local tag = "DIDL%-Lite"
    local val1, val2, val3 = xml1:match("(<"..tag.." ?.->)(.*)(</"..tag..">)")
    local val4, val5, val6 = xml2:match("(<"..tag.." ?.->)(.*)(</"..tag..">)")
    if (val1 and val2 and val3 and val4 and val5 and val6) then
        result = val1 .. val2 .. val5 .. val3
    elseif (val1 and val2 and val3) then
        result = xml1
    elseif (val4 and val5 and val6) then
        result = xml2
    end
    return result
  end


  function browseContent(uuid, serviceType, browseObj, onlyMetadata)
      local result = ""
      local ContentDirectory = getService(uuid, serviceType)
      if (ContentDirectory == nil) then
          return result
      end

      local browseFlag
      if (onlyMetadata) then
          browseFlag = "BrowseMetadata"
      else
          browseFlag = "BrowseDirectChildren"
      end

      local fetched = 0
      local total = 0
      local status, tmp
      repeat
          status, tmp = ContentDirectory.Browse({OrderedArgs={
                                 "ObjectID=" .. browseObj,
                                 "BrowseFlag=" .. browseFlag,
                                 "Filter=",
                                 "StartingIndex=" .. string.format("%d", fetched),
                                 "RequestedCount=100",
                                 "SortCriteria="}})
          if (status == true) then
              result = mergeXml(result, extractElement("Result", tmp, ""))
              fetched = fetched + tonumber(extractElement("NumberReturned", tmp, ""))
              total = tonumber(extractElement("TotalMatches", tmp, ""))
          else
              total = 0
          end
      until fetched >= total
      return result
  end


  --------------------------------------------------------------------------------
  -- UPnP Event Proxy
  --------------------------------------------------------------------------------


  -- XML pattern used when notifying the UPnP event proxy of the the subscription to an event
  local PROXY_REQUEST = "<subscription expiry='%d'><variable name='%s' host='localhost' deviceId='%d' serviceId='%s' action='%s' parameter='%s' sidParameter='sid'/></subscription>"

  -- Version of the UPnP event proxy, or nil if proxy is not running or not used
  local ProxyApiVersion = nil


  -- getProxyApiVersion()
  -- Calls the proxy with GET /version.
  -- Sets the ProxyApiVersion global variable to the value received
  -- Return value:
  --   nil if the proxy is not running.
  --   The proxy API version (as a string) otherwise.
  function getProxyApiVersion()
    local sock = function()
        local s = socket.tcp()
        s:settimeout(2)
        return s
    end

    local t = {}
    local request, code = http.request({
        url = "http://localhost:2529/version",
        create = sock,
        sink = ltn12.sink.table(t)
    })

    if (request == nil and code == "timeout") then
        -- Proxy may be busy.
        error("Temporarily cannot communicate with proxy", 2)
        return nil
    elseif (request == nil and code ~= "closed") then
        -- Proxy not running.
        error("Cannot contact UPnP event proxy: " .. code, 1)
        return ""
    else
        -- Proxy is running, note its version number.
        ProxyApiVersion = table.concat(t)
        return ProxyApiVersion
    end
  end


  -- WeMo plugin contrib
  -- proxyVersionAtLeast(n)
  -- Returns true if the proxy is running and is at least version n.
  function proxyVersionAtLeast(n)
    if (ProxyApiVersion and tonumber(ProxyApiVersion:match("^(%d+)")) >= n) then
        return true
    end
    return false
  end


  -- unuseProxy()
  -- Disable the usage of the UPnP event proxy by resetting the variable ProxyApiVersion to nil
  -- Return value: none
  function unuseProxy()
    ProxyApiVersion = nil
    log("UPnP event proxy is now unused", 2)
  end


  -- subscribeToUPnPEvent(device, ip, port, serverURL, eventVariable, actionServiceId, actionName, renewalSID)
  -- Process a new subscription to an event for an UPnP device, or renew an active subscription
  -- First send a subscription request to the UPnP device and then notify the proxy of this new subscription
  -- Parameters:
  --   device is the device id that the UPnP event proxy will notify
  --   ip is the IP address of the UPnP device
  --   port is the port to be used to communicate with the UPnP device
  --   serverURL is the URL to be used to subscribe to the UPnP device
  --   eventVariable is the UPnP variable name we subscribe to
  --   actionServiceId is the service id of the action to be called by the UPnP event proxy
  --   actionName is the action name to be called by the UPnP event proxy
  --   renewalSID is the subscription id to be renewed or nil if it is an initial subscription
  -- Return values:
  --   subscription id or nil if the subscribe process failed
  --   expiry date of the new or renewed subscription; nil if the subscribe process failed
  --   live duration of the new or renewed subscription; nil if the subscribe process failed 
  function subscribeToUPnPEvent(device, ip, port, serverURL, eventVariable, actionServiceId, actionName, renewalSID)
    log("Subscribing to event " .. serverURL .. " for variable " .. eventVariable .. " with SID " .. (renewalSID or "nil"))

	local sock = function()
		local s = socket.tcp()
		s:settimeout(5)
		return s
	end

	-- Ask the device to inform the proxy about status changes.
	local expiry = nil
	local sid, duration = UPnP_subscribe(ip, port, serverURL, "http://" .. VeraIP .. ":2529/upnp/event", renewalSID)
	if (sid) then
		expiry = os.time() + duration

		-- Tell proxy about this subscription and the variable we care about.
		-- Volume is the variable we care about.
		local proxyRequestBody = PROXY_REQUEST:format(expiry, eventVariable, device, actionServiceId, actionName, eventVariable)
		local request, code = http.request({
			url = "http://localhost:2529/upnp/event/" .. url.escape(sid),
			create = sock,
			method = "PUT",
			headers = {
				["Content-Type"] = "text/xml",
				["Content-Length"] = proxyRequestBody:len(),
			},
			source = ltn12.source.string(proxyRequestBody),
			sink = ltn12.sink.null(),
		})
-- TODO retry if timeout
		if (request == nil and code ~= "closed") then
			error("Failed to notify proxy of subscription: " .. code)
			sid = nil
			expiry = nil
		elseif (code ~= 200) then
			error("Failed to notify proxy of subscription: " .. code)
			sid = nil
			expiry = nil
		else
			log("Successfully notified proxy of subscription")
		end
	end

	return sid, expiry, duration
  end


  -- WeMo plugin contrib
  -- cancelProxySubscription(sid)
  -- Sends a DELETE /upnp/event/[sid] message to the UPnP event proxy,
  -- Return value:
  --   nil if the proxy timed out (should try again).
  --   false if the proxy refused our request (permanently).
  --   true if the proxy agreed to our request.
  function cancelProxySubscription(sid)
    log("Cancelling subscription for sid " .. sid)
    local sock = function()
        local s = socket.tcp()
        s:settimeout(2)
        return s
    end

    local request, code = http.request({
        url = "http://localhost:2529/upnp/event/" .. url.escape(sid),
        create = sock,
        method = "DELETE",
        source = ltn12.source.empty(),
        sink = ltn12.sink.null(),
    })
    if (request == nil and code ~= "closed") then
        error("Failed to cancel subscription: " .. code)
        return nil
    elseif (code ~= 200) then
        error("Failed to cancel subscription: " .. code)
        return false
    else
        log("Successfully cancelled subscription")
        return true
    end
  end


  -- subscribeToEvents(device, ip, port, subscriptions, actionServiceId, uuid)
  -- Process the subscription to several events for an UPnP device, or renew these subscriptions
  -- The subscriptions are defined in a table (subscriptions).
  -- Each element of this table is enhanced with subsciption data (id, expiry date, ...).
  -- Before calling this function, the UPnP services for the UPnP device have to be loaded. Only
  -- events associated to a loaded and registered service will be subscribed to.
  -- If one event subscription fails, all subscriptions are finally cancelled.
  -- A timer is set to call the function renewSubscriptions, so that all the subscritopns can be
  -- renewed in time. So this function "renewSubscriptions" must be declared by the plugin.
  -- Parameters:
  --   device is the device id that the UPnP event proxy will notify
  --   ip is the IP address of the UPnP device
  --   port is the port to be used to communicate with the UPnP device
  --   subscriptions is a table defining the events to be subscribed to
  --   actionServiceId is the service id for all actions to be called by the UPnP event proxy
  --   uuid is the UUID of the UPnP device
  -- Return value:
  --   true if UPnP event proxy is not running or not used
  --   true if the UPnP services are not yet loaded and registered
  --   true if all subscriptions succeeded
  --   false if subscriptions failed
  function subscribeToEvents(device, ip, port, subscriptions, actionServiceId, uuid)
    log("Subscribing to all events")

    if (proxyVersionAtLeast(1) == false) then
        return true
    end
    if (aresServicesLoaded(uuid) == false) then
        return true
    end

    local result = true

    local sid, expiry, duration

    local nbSubscriptions = 0
    local minDuration = 3600

    for i,subscription in ipairs(subscriptions) do
        if (Services[uuid][subscription.service] ~= nil) then
            sid = nil
            if (subscription.id ~= "") then
                sid = subscription.id
            end
            sid, expiry, duration = subscribeToUPnPEvent(device,
                                                         ip,
                                                         port,
                                                         Services[uuid][subscription.service].eventSubURL,
                                                         subscription.eventVariable,
                                                         actionServiceId,
                                                         subscription.actionName,
                                                         sid)
            if (sid ~= nil) then
                log("Event subscription succeeded => SID " .. sid .. " duration " .. duration .. " expiry " .. expiry)
                subscription.id = sid
                subscription.expiry = expiry
                nbSubscriptions = nbSubscriptions + 1
                if (duration < minDuration) then
                    minDuration = duration
                end
            else
                error("Event subscription failed: " .. Services[uuid][subscription.service].eventSubURL)
                cancelProxySubscriptions(subscriptions)
                nbSubscriptions = 0
                result = false
                break
            end
        end
    end

    if (nbSubscriptions > 0) then
        if (minDuration >= 600) then
            minDuration = minDuration - 300
        end

        luup.call_delay("renewSubscriptions", minDuration, device .. ":" .. uuid)
    end

    return result
  end


  -- cancelProxySubscriptions(subscriptions)
  -- Sends a DELETE /upnp/event/[sid] message to the UPnP event proxy for all subscriptions
  -- defined in the table subscriptions.
  -- Subscription data for each element of this table are reset.
  -- Parameters:
  --   subscriptions is a table containing the current subscription data
  -- Return value: none
  function cancelProxySubscriptions(subscriptions)
    log("Cancelling all event subscriptions")

    if (proxyVersionAtLeast(1) == false) then
        return
    end

    for i,subscription in ipairs(subscriptions) do
        if (subscription.id ~= "") then
            cancelProxySubscription(subscription.id)
            subscription.id = ""
            subscription.expiry = ""
        end
    end
  end


  -- isValidNotification(notifyAction, sid, subscriptions)
  -- Check whether a proxy notification is valid (the subscription id has to be declared in the
  -- table containing current subscription data)
  -- If not, a delayed action cancelProxySubscription is called to unsubscribe this subscription.
  -- So this function "cancelProxySubscription" must be declared by the plugin.
  -- Parameters:
  --   notifyAction is the action called by the UPnP event proxy
  --   sid is the subscription id provided by the UPnP event proxy
  --   subscriptions is a table containing the current subscription data
  -- Return value:
  --   true if the proxy notification is valid
  --   false if the proxy notification is not valid
  function isValidNotification(notifyAction, sid, subscriptions)
    local valid = false
    if (proxyVersionAtLeast(1) == false) then
        error("Call to " .. notifyAction .. " while proxy is not used", 2)
    else
        for i,subscription in ipairs(subscriptions) do
            if (subscription.id == sid) then
                valid = true
                break
            end
        end
        if (not valid) then
            error("Call to " .. notifyAction .. " with bad SID " .. sid, 2)
        end
    end
    if (not valid) then
        -- Try to shut the proxy up, we don't care about this SID.
        luup.call_delay("cancelProxySubscription", 1, sid)
    end
    return valid
  end
